function [ind1,ind2,A,b]=wincaalign(x1,w1,x2,w2,wd,Nitmax,radius,Nblocks)
%	INCA ALIGNMENT
%
%	[ind1,ind2]=wincaalign(x1,w1,x2,w2,wd,Nitmax,radius,Nblocks)
%
%	x1: 1st set of N1 D-dimensional column vectors (TARGET)
%	w1: relative weights of vectors in x1 (size 1xN1)
%	x2: 2nd set of N2 D-dimensional column vectors (SOURCE)
%	w2: relative weights of vectors in x2 (size 1xN2)
%	wd: weights for squared Euclidean distance (size Dx1, def.: ones)
%	Nitmax: maximum number of INCA iterations (def.: 25)
%	radius: number of upper/lower bands of auxiliary linear transform (def.: full)
%	Nblocks: number of blocks of auxiliary linear transform (def.: 1)
%
%	ind1, ind2: indices of pairs (with repetitions), x1(:,ind1) <-> x2(:,ind2)
%
%	Warning: please note that x1 is the target and x2 is the source!!
%
%	---------------------------------------------------
%	Copyright (c) 2015 University of the Basque Country
%	---------------------------------------------------
%	License
%	This program is free software: you can redistribute
%	it and/or modify it under the terms of the GNU General
%	Public License as published by the Free Software Foundation,
%	either version 3 of the License, or any later version.
%	This program is distributed in the hope that it will be
%	useful, but WITHOUT ANY WARRANTY; without even the implied
%	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
%	PURPOSE. See the GNU General Public License for more details.
%	---------------------------------------------------
%	Author: Daniel Erro (derro(a)aholab.ehu.es)
%	---------------------------------------------------


if nargin<4 error('Missing args.'); end;
% default values of optional variables
if nargin<5 wd=[]; end;
if nargin<6 || isempty(Nitmax) Nitmax=25; end;
if nargin<7 || isempty(radius) radius=inf; end;
if nargin<8 || isempty(Nblocks) Nblocks=1; end;
% get dimension and number of vectors in set 1
[dim,N1]=size(x1);
% by default, same weight to all vectors, and normalize just in case
if isempty(w1) w1=repmat(1/N1,1,N1); else w1=w1./sum(w1); end;
% get number of vectors in set 2
N2=size(x2,2);
% by default, same weight to all vectors, and normalize just in case
if isempty(w2) w2=repmat(1/N2,1,N2); else w2=w2./sum(w2); end;
% check dimension compatibility
if size(x2,1)~=dim error('Incompatible dimensions.'); end;
% check/clip optional variables
if Nitmax<=0 Nitmax=1; end;
if Nblocks<=0 || rem(dim,Nblocks)~=0 error('Incorrect number of blocks.'); end;
radius=min(max(radius,0),dim/Nblocks-1);
% initialize transform (identity) and previous average nn distance (inf)
A=eye(dim);
b=zeros(dim,1);
sda=realmax;
% let's go
fprintf('INCA-START\n');
% bi-directional nearest neighbor search
[ind1,ind2,sd]=nnpairs_(x1,w1,x2,w2,wd);
% iterations
it=1;
while sd<sda && it<=Nitmax
    fprintf('Iter %d: avg nn distance %.6f\n',it,sd);
    % creating backups
    ind1a=ind1; ind2a=ind2; sda=sd; Aa=A; ba=b;
    % train linear transform that projects set 2 onto set 1
	[A,b]=blocklintransf_(x1(:,ind1),x2(:,ind2),[w1 w2],wd,radius,Nblocks);
	% bi-directional nearest neighbor search
	[ind1,ind2,sd]=nnpairs_(x1,w1,A*x2+repmat(b,1,N2),w2,wd);
	% increase iteration number
	it=it+1;
end;
% finished
fprintf('INCA-END\n');
% if nn distance has increased, restore backups
if sd>sda ind1=ind1a; ind2=ind2a; A=Aa; b=ba; end;

% nearest neighbor pairs in both directions
function [ind1,ind2,sumd]=nnpairs_(x1,w1,x2,w2,wd)
% get dimensions (already checked)
[dim,N1]=size(x1);
N2=size(x2,2);
% initialize with zeros
ind1=zeros(1,N1);
ind2=zeros(1,N2);
sumd=0.0;
% normalize weights just in case
aux=sum(w1);
if aux~=1 w1=w1./aux; end;
aux=sum(w2);
if aux~=1 w2=w2./aux; end;
% let's go
d2=repmat(realmax,1,N2);
for k1=1:N1
    d1=realmax;
    for k2=1:N2
		% weighted squared Eculidean distance between x1(:,k1) and x2(:,k2)
        dk1k2=(x1(:,k1)-x2(:,k2)).^2;
		if ~isempty(wd) dk1k2=wd.*dk1k2; end;
		dk1k2=sum(dk1k2);
		% update nearest neighbor of x1(:,k1) if necessary
        if dk1k2<d1 d1=dk1k2; ind1(k1)=k2; end;
		% update nearest neighbor of x2(:,k2) if necessary
        if dk1k2<d2(k2) d2(k2)=dk1k2; ind2(k2)=k1; end;
    end;
    sumd=sumd+w1(k1)*d1;
end;
% average nn distance
sumd=0.5*(sumd+sum(w2.*d2));
% re-structure ind1 and ind2
d1=ind1; d2=ind2;
ind1=[1:N1 d2];
ind2=[d1 1:N2];
d1=[]; d2=[];

% train blockwise linear transform that projects x2 onto x1
function [A,b]=blocklintransf_(x1,x2,w,wd,radius,Nblocks)
% get dimension and allocate transform
dim=size(x1,1);
A=zeros(dim);
b=zeros(dim,1);
% block dimension
bdim=floor(dim/Nblocks);
% indices of current block
ind=1:bdim;
% solve on a block-by-block basis
for k=1:Nblocks
    % solve for current block
    [A(ind,ind),b(ind)]=lintransf_(x1(ind,:),x2(ind,:),w,wd,radius);
    % indices of next block
    ind=ind+bdim;
end;

% train linear transform that projects x2 onto x1
function [A,b]=lintransf_(x1,x2,w,wd,radius)
% get dimensions (assuming size(x1)=size(x2)
[dim,N]=size(x1);
% append ones to x2 to solve for A'=[A b]
x2=[x2; ones(1,N)];
% allocate matrix
A=zeros(dim,dim+1);
% solve on a row-by-row basis
for k=1:dim
	% indices of non-zero parts of row k according to radius
    ind=[max(1,k-radius):min(k+radius,dim) dim+1];
	% solve for row k
	if isempty(wd)
		term=(w.*x1(k,:))*x2(ind,:).';
		Term=(repmat(w,length(ind),1).*x2(ind,:))*x2(ind,:).';
	else
		term=(wd(k)*w.*x1(k,:))*x2(ind,:).';
		Term=(repmat(wd(k)*w,length(ind),1).*x2(ind,:))*x2(ind,:).';
	end;
    A(k,ind)=term/Term;
end;
% separate matrix and vector
b=A(:,dim+1); A=A(:,1:dim);
